@@ -23,10 +23,8 @@ from __future__ import print_function, division, absolute_import
 
 import numpy as np
 import platform
-import sys
 from os.path import join, dirname, isfile, realpath
 import os
-import functools
 import warnings
 
 import ctypes
@@ -40,6 +38,10 @@ class mxArray(ctypes.Structure):
 class Engine(ctypes.Structure):
     pass
 
+mwSize = c_size_t
+mwIndex = c_size_t
+
+
 wrap_script = r"""
 ERRSTR__ = '';
 try
@@ -90,15 +92,8 @@ class MatlabSession(object):
     def __init__(self, options='-nosplash', matlab_root=None, buffer_size=0):
         system = platform.system()
 
-
-        ### Find MATLAB's root path
-        if matlab_root is None:
-            path_dirs = os.environ.get("PATH").split(os.pathsep)
-            for path_dir in path_dirs:
-                candidate = realpath(join(path_dir, 'matlab'))
-                if isfile(candidate):
-                    matlab_root = dirname(dirname(candidate))
-                    break
+        matlab_root = find_matlab_root()
+        lib_dir = find_lib_dir(matlab_root)
 
         if matlab_root is None:
             raise RuntimeError("MATLAB location is unknown (try to initialize MatlabSession with matlab_root properly set)")
@@ -106,29 +101,37 @@ class MatlabSession(object):
         self._matlab_root = matlab_root
 
 
-        ### Load libraries and start engine
-        if system in ('Linux', 'Darwin'):
+        ### Load libraries the libraries
+        if system == 'Linux':
             self._libeng = ctypes.CDLL(
-                join(matlab_root, 'bin', 'glnxa64', 'libeng.so')
+                join(lib_dir, 'libeng.so')
             )
-            self._libmx = ctypes.CDLL(
-                join(matlab_root, 'bin', 'glnxa64', 'libmx.so')
+            self._libmx = Library(
+                join(lib_dir, 'libmx.so')
             )
-            executable = join(matlab_root, 'bin', 'matlab')
-            command = "{} {}".format(executable, options)
 
-        # elif system=='Windows':
-        #     self.engine = CDLL(join(matlab_root,'bin','glnxa64','libeng.dll'))
-        #     self.mx = CDLL(join(matlab_root,'bin','glnxa64','libmx.dll'))
-        #     self.ep = self.engine.engOpen(None)
+            command = "{executable} {options}".format(
+                executable=join(matlab_root, 'bin', 'matlab'),
+                options=options
+            )
+
+        elif system=='Windows':
+            if lib_dir not in os.environ['PATH']:
+                os.environ['PATH'] = lib_dir + ';' + os.environ['PATH']
+
+            self._libeng = ctypes.CDLL('libeng')
+            self._libmx = Library('libmx')
+
+            command = None
 
         else:
-            raise NotImplementedError("System {} not supported".format(system))
+            raise NotImplementedError("System {} not yet supported.".format(system))
 
 
 
 
         ### Setup function types for args and returns
+        ## libeng
         self._libeng.engOpen.argtypes = (c_char_p,)
         self._libeng.engOpen.restype = POINTER(Engine)
         self._libeng.engOpen.errcheck = error_check
@@ -149,11 +152,12 @@ class MatlabSession(object):
         self._libeng.engOutputBuffer.restype = c_int
 
 
+        ## libmx
         self._libmx.mxGetNumberOfDimensions.argtypes = (POINTER(mxArray),)
-        self._libmx.mxGetNumberOfDimensions.restype = c_size_t
+        self._libmx.mxGetNumberOfDimensions.restype = mwSize
 
         self._libmx.mxGetDimensions.argtypes = (POINTER(mxArray),)
-        self._libmx.mxGetDimensions.restype = POINTER(c_size_t)
+        self._libmx.mxGetDimensions.restype = POINTER(mwSize)
 
         self._libmx.mxGetNumberOfElements.argtypes = (POINTER(mxArray),)
         self._libmx.mxGetNumberOfElements.restype = c_size_t
@@ -181,12 +185,12 @@ class MatlabSession(object):
         self._libmx.mxGetImagData.restype = POINTER(c_void_p)
         self._libmx.mxGetImagData.errcheck = error_check
 
-        self._libmx.mxGetCell.argtypes = (POINTER(mxArray), c_size_t)
+        self._libmx.mxGetCell.argtypes = (POINTER(mxArray), mwIndex)
         self._libmx.mxGetCell.restype = POINTER(mxArray)
         ### Errors has to be handled elswhere, because of NULL on uninitialized cells
         # self._libmx.mxGetCell.errcheck = error_check
 
-        self._libmx.mxSetCell.argtypes = (POINTER(mxArray), c_size_t, POINTER(mxArray))
+        self._libmx.mxSetCell.argtypes = (POINTER(mxArray), mwIndex, POINTER(mxArray))
         self._libmx.mxSetCell.restype = None
 
         self._libmx.mxGetNumberOfFields.argtypes = (POINTER(mxArray),)
@@ -197,16 +201,15 @@ class MatlabSession(object):
         self._libmx.mxGetFieldNameByNumber.restype = c_char_p
         self._libmx.mxGetFieldNameByNumber.errcheck = error_check
 
-        self._libmx.mxGetField.argtypes = (POINTER(mxArray), c_size_t, c_char_p)
+        self._libmx.mxGetField.argtypes = (POINTER(mxArray), mwIndex, c_char_p)
         self._libmx.mxGetField.restype = POINTER(mxArray)
         ### Errors has to be handled elswhere, because of NULL on uninitialized fields
         # self._libmx.mxGetField.errcheck = error_check
 
-        # void mxSetField(mxArray *pm, mwIndex index, const char *fieldname, mxArray *pvalue);
-        self._libmx.mxSetField.argtypes = (POINTER(mxArray), c_size_t, c_char_p, POINTER(mxArray))
+        self._libmx.mxSetField.argtypes = (POINTER(mxArray), mwIndex, c_char_p, POINTER(mxArray))
         self._libmx.mxSetField.restype = None
 
-        self._libmx.mxCreateStructArray.argtypes = (c_size_t, POINTER(c_size_t), c_int, POINTER(c_char_p))
+        self._libmx.mxCreateStructArray.argtypes = (mwSize, POINTER(mwSize), c_int, POINTER(c_char_p))
         self._libmx.mxCreateStructArray.restype = POINTER(mxArray)
         self._libmx.mxCreateStructArray.errcheck = error_check
 
@@ -218,19 +221,19 @@ class MatlabSession(object):
         self._libmx.mxCreateString.restype = POINTER(mxArray)
         self._libmx.mxCreateString.errcheck = error_check
 
-        self._libmx.mxGetString.argtypes = (POINTER(mxArray), c_char_p, c_size_t)
+        self._libmx.mxGetString.argtypes = (POINTER(mxArray), c_char_p, mwSize)
         self._libmx.mxGetString.restype = c_int
         self._libmx.mxGetString.errcheck = error_check
 
-        self._libmx.mxCreateNumericArray.argtypes = (c_size_t, POINTER(c_size_t), c_int, c_int)
+        self._libmx.mxCreateNumericArray.argtypes = (mwSize, POINTER(mwSize), c_int, c_int)
         self._libmx.mxCreateNumericArray.restype = POINTER(mxArray)
         self._libmx.mxCreateNumericArray.errcheck = error_check
 
-        self._libmx.mxCreateLogicalArray.argtypes = (c_size_t, POINTER(c_size_t))
+        self._libmx.mxCreateLogicalArray.argtypes = (mwSize, POINTER(mwSize))
         self._libmx.mxCreateLogicalArray.restype = POINTER(mxArray)
         self._libmx.mxCreateLogicalArray.errcheck = error_check
 
-        self._libmx.mxCreateCellArray.argtypes = (c_size_t, POINTER(c_size_t))
+        self._libmx.mxCreateCellArray.argtypes = (mwSize, POINTER(mwSize))
         self._libmx.mxCreateCellArray.restype = POINTER(mxArray)
         self._libmx.mxCreateCellArray.errcheck = error_check
 
@@ -260,8 +263,8 @@ class MatlabSession(object):
 
         ### Check MATLAB version
         version = self.version
-        if 'R2014b' in version:
-            warnings.warn("You are using MATLAB version R2014a, which appears to have a bug in engGetVariable().  You will only be able to get arrays of type double.")
+        if ('R2014a' in version) and system == 'Linux':
+            warnings.warn("You are using MATLAB version R2014a on Linux, which appears to have a bug in engGetVariable().  You will only be able to use arrays of type double.")
 
 
 
@@ -298,10 +301,6 @@ class MatlabSession(object):
 
         error_string = self._libmx.mxArrayToString(mxresult)
 
-        # TODO: possible memory leak, error_string should be freed
-        # (alternatively use mxGetString), perhaps error_string is
-        # handled by python (?)
-
         if error_string != "":
             raise RuntimeError("Error from MATLAB\n{0}".format(error_string))
 
@@ -323,17 +322,6 @@ class MatlabSession(object):
         """
         pm = self._libeng.engGetVariable(self._ep, name)
 
-        ndims = self._libmx.mxGetNumberOfDimensions(pm)
-        dims = self._libmx.mxGetDimensions(pm)
-        numelems = self._libmx.mxGetNumberOfElements(pm)
-        elem_size = self._libmx.mxGetElementSize(pm)
-        class_name = self._libmx.mxGetClassName(pm)
-        is_numeric = self._libmx.mxIsNumeric(pm)
-        is_complex = self._libmx.mxIsComplex(pm)
-        data = self._libmx.mxGetData(pm)
-        imag_data = self._libmx.mxGetImagData(pm)
-
-
         out = mxarray_to_ndarray(self._libmx, pm)
 
 
@@ -387,6 +375,44 @@ def error_check(result, func, arguments):
 
 
 
+def find_matlab_root():
+    """Look for matlab binary and return root directory of MATLAB
+    installation.
+
+    """
+    matlab_root = None
+
+    path_dirs = os.environ.get("PATH").split(os.pathsep)
+    for path_dir in path_dirs:
+        candidate = realpath(join(path_dir, 'matlab'))
+        if isfile(candidate) or isfile(candidate + '.exe'):
+            matlab_root = dirname(dirname(candidate))
+            break
+
+    return matlab_root
+
+
+def find_lib_dir(matlab_root):
+    """Locate the directory where libraries are located, which is OS and
+    architecture dependent.
+
+    """
+    bits, linkage = platform.architecture()
+    system = platform.system()
+
+    if (system == 'Linux') and (bits == '64bit'):
+        lib_dir = join(matlab_root, "bin", "glnxa64")
+    # elif (system == 'Linux') and (bits == '32bit'):
+    #     lib_dir = join(matlab_root, "bin", "glnx86")
+    elif (system == 'Windows') and (bits == '64bit'):
+        lib_dir = join(matlab_root, "bin", "win64")
+    # elif (system == 'Windows') and (bits == '32bit'):
+    #     lib_dir = join(matlab_root, "bin", "win32")
+    else:
+        raise RuntimeError("Unsopported OS or architecture: {} {}".format(system, bits))
+
+    return lib_dir
+
 
 class Workspace(object):
     def __init__(self, session):
@@ -411,7 +437,7 @@ class Workspace(object):
             out = MatlabFunction(name=attr, session=self._session)
 
         else:
-            raise NotImplemented("Unknown variable/function type in MATLAB workspace: {}".format(attr))
+            raise NotImplementedError("Unknown variable/function type in MATLAB workspace: {}".format(attr))
 
         return out
 
@@ -669,7 +695,7 @@ def ndarray_to_mxarray(libmx, arr):
 
 
     if isinstance(arr, np.ndarray) and arr.dtype.kind in ['i','u','f','c']:
-        dim = arr.ctypes.shape_as(c_size_t)
+        dim = arr.ctypes.shape_as(mwSize)
         complex_flag = (arr.dtype.kind == 'c')
 
         pm = libmx.mxCreateNumericArray(
@@ -690,7 +716,7 @@ def ndarray_to_mxarray(libmx, arr):
 
 
     elif isinstance(arr, np.ndarray) and arr.dtype.kind == 'b':
-        dim = arr.ctypes.shape_as(c_size_t)
+        dim = arr.ctypes.shape_as(mwSize)
 
         pm = libmx.mxCreateLogicalArray(arr.ndim, dim)
 
@@ -700,7 +726,7 @@ def ndarray_to_mxarray(libmx, arr):
 
 
     elif isinstance(arr, np.ndarray) and arr.dtype.kind in ('O', 'S'):
-        dim = arr.ctypes.shape_as(c_size_t)
+        dim = arr.ctypes.shape_as(mwSize)
 
         pm = libmx.mxCreateCellArray(arr.ndim, dim)
 
@@ -710,7 +736,7 @@ def ndarray_to_mxarray(libmx, arr):
 
 
     elif isinstance(arr, np.ndarray) and len(arr.dtype) > 0:
-        dim = arr.ctypes.shape_as(c_size_t)
+        dim = arr.ctypes.shape_as(mwSize)
 
         name_num = len(arr.dtype.names)
 
@@ -734,3 +760,27 @@ def ndarray_to_mxarray(libmx, arr):
         raise NotImplementedError('Type {} not supported.'.format(arr.dtype))
 
     return pm
+
+
+
+class Library(object):
+    """Shared library proxy.
+
+    The purpouse of this class is to wrap CDLL objects and append
+    `_730` to function names on the fly.  It should resolve the int vs
+    mwSize problems for those functions.
+
+    """
+    def __init__(self, *args, **kwargs):
+        self._lib = ctypes.CDLL(*args, **kwargs)
+
+    def __getattr__(self, attr):
+
+        attr730 = attr + '_730'
+
+        try:
+            out = getattr(self._lib, attr730)
+        except AttributeError:
+            out = getattr(self._lib, attr)
+
+        return out
